@@ -1,7 +1,7 @@
 import ast
 import keyword
 import re
-from typing import TYPE_CHECKING, Iterable
+from typing import TYPE_CHECKING, Any, Dict, Iterable, List
 
 from prompt_toolkit.completion import (
     CompleteEvent,
@@ -129,7 +129,10 @@ class PythonCompleter(Completer):
             for c in self.dictionary_completer.get_completions(
                 document, complete_event
             ):
-                has_dict_completions = True
+                if c.text not in "[.":
+                    # If we get the [ or . completion, still include the other
+                    # completions.
+                    has_dict_completions = True
                 yield c
             if has_dict_completions:
                 return
@@ -206,8 +209,8 @@ class DictionaryCompleter(Completer):
     """
     Experimental completer for Python dictionary keys.
 
-    Warning: This does an `eval` on the Python object before the open square
-             bracket, which is potentially dangerous. It doesn't match on
+    Warning: This does an `eval` and `repr` on some Python expressions before
+             the cursor, which is potentially dangerous. It doesn't match on
              function calls, so it only triggers attribute access.
     """
 
@@ -217,53 +220,160 @@ class DictionaryCompleter(Completer):
         self.get_globals = get_globals
         self.get_locals = get_locals
 
-        self.pattern = re.compile(
-            r"""
-                # Any expression safe enough to eval while typing.
-                # No operators, except dot, and only other dict lookups.
-                # Technically, this can be unsafe of course, if bad code runs
-                # in `__getattr__` or ``__getitem__``.
-                (
-                    # Variable name
-                    [a-zA-Z0-9_]+
+        # Pattern for expressions that are "safe" to eval for auto-completion.
+        # These are expressions that contain only attribute and index lookups.
+        varname = r"[a-zA-Z_][a-zA-Z0-9_]*"
 
-                    \s*
+        expression = rf"""
+            # Any expression safe enough to eval while typing.
+            # No operators, except dot, and only other dict lookups.
+            # Technically, this can be unsafe of course, if bad code runs
+            # in `__getattr__` or ``__getitem__``.
+            (
+                # Variable name
+                {varname}
 
-                    (?:
-                        # Attribute access.
-                        \s* \. \s* [a-zA-Z0-9_]+ \s*
+                \s*
 
-                        |
+                (?:
+                    # Attribute access.
+                    \s* \. \s* {varname} \s*
 
-                        # Item lookup.
-                        # (We match the square brackets. We don't care about
-                        # matching quotes here in the regex. Nested square
-                        # brackets are not supported.)
-                        \s* \[ [a-zA-Z0-9_'"\s]+ \] \s*
-                    )*
-                )
+                    |
 
-                # Dict loopup to complete (square bracket open + start of
-                # string).
-                \[
-                \s* ([a-zA-Z0-9_'"]*)$
+                    # Item lookup.
+                    # (We match the square brackets. The key can be anything.
+                    # We don't care about matching quotes here in the regex.
+                    # Nested square brackets are not supported.)
+                    \s* \[ [^\[\]]+ \] \s*
+                )*
+            )
+        """
+
+        # Pattern for recognizing for-loops, so that we can provide
+        # autocompletion on the iterator of the for-loop. (According to the
+        # first item of the collection we're iterating over.)
+        self.for_loop_pattern = re.compile(
+            rf"""
+                for \s+ ([a-zA-Z0-9_]+) \s+ in \s+ {expression} \s* :
             """,
             re.VERBOSE,
         )
 
+        # Pattern for matching a simple expression (for completing [ or .
+        # operators).
+        self.expression_pattern = re.compile(
+            rf"""
+                {expression}
+                $
+            """,
+            re.VERBOSE,
+        )
+
+        # Pattern for matching item lookups.
+        self.item_lookup_pattern = re.compile(
+            rf"""
+                {expression}
+
+                # Dict loopup to complete (square bracket open + start of
+                # string).
+                \[
+                \s* ([^\[\]]*)$
+            """,
+            re.VERBOSE,
+        )
+
+        # Pattern for matching attribute lookups.
+        self.attribute_lookup_pattern = re.compile(
+            rf"""
+                {expression}
+
+                # Attribute loopup to complete (dot + varname).
+                \.
+                \s* ([a-zA-Z0-9_]*)$
+            """,
+            re.VERBOSE,
+        )
+
+    def _lookup(self, expression: str, temp_locals: Dict[str, Any]) -> object:
+        """
+        Do lookup of `object_var` in the context.
+        `temp_locals` is a dictionary, used for the locals.
+        """
+        try:
+            return eval(expression.strip(), self.get_globals(), temp_locals)
+        except BaseException:
+            return None  # Many exception, like NameError can be thrown here.
+
     def get_completions(
         self, document: Document, complete_event: CompleteEvent
     ) -> Iterable[Completion]:
-        match = self.pattern.search(document.text_before_cursor)
+
+        # First, find all for-loops, and assing the first item of the
+        # collections they're iterating to the iterator variable, so that we
+        # can provide code completion on the iterators.
+        temp_locals = self.get_locals().copy()
+
+        for match in self.for_loop_pattern.finditer(document.text_before_cursor):
+            varname, expression = match.groups()
+            expression_val = self._lookup(expression, temp_locals)
+
+            # We do this only for lists and tuples. Calling `next()` on any
+            # collection would create undesired side effects.
+            if isinstance(expression_val, (list, tuple)) and expression_val:
+                temp_locals[varname] = expression_val[0]
+
+        # Get all completions.
+        yield from self._get_expression_completions(
+            document, complete_event, temp_locals
+        )
+        yield from self._get_item_lookup_completions(
+            document, complete_event, temp_locals
+        )
+        yield from self._get_attribute_completions(
+            document, complete_event, temp_locals
+        )
+
+    def _do_repr(self, obj: object) -> str:
+        try:
+            return str(repr(obj))
+        except BaseException:
+            raise ReprFailedError
+
+    def _get_expression_completions(
+        self,
+        document: Document,
+        complete_event: CompleteEvent,
+        temp_locals: Dict[str, Any],
+    ) -> Iterable[Completion]:
+        """
+        Complete the [ or . operator after an object.
+        """
+        match = self.expression_pattern.search(document.text_before_cursor)
+        if match is not None:
+            object_var = match.groups()[0]
+            result = self._lookup(object_var, temp_locals)
+
+            if isinstance(result, (list, tuple, dict)):
+                yield Completion("[", 0)
+            elif result:
+                yield Completion(".", 0)
+
+    def _get_item_lookup_completions(
+        self,
+        document: Document,
+        complete_event: CompleteEvent,
+        temp_locals: Dict[str, Any],
+    ) -> Iterable[Completion]:
+        """
+        Complete dictionary keys.
+        """
+        match = self.item_lookup_pattern.search(document.text_before_cursor)
         if match is not None:
             object_var, key = match.groups()
-            object_var = object_var.strip()
 
             # Do lookup of `object_var` in the context.
-            try:
-                result = eval(object_var, self.get_globals(), self.get_locals())
-            except BaseException:
-                return  # Many exception, like NameError can be thrown here.
+            result = self._lookup(object_var, temp_locals)
 
             # If this object is a dictionary, complete the keys.
             if isinstance(result, dict):
@@ -279,7 +389,75 @@ class DictionaryCompleter(Completer):
 
                 for k in result:
                     if str(k).startswith(key_obj):
-                        yield Completion(str(repr(k)), -len(key), display=str(repr(k)))
+                        try:
+                            k_repr = self._do_repr(k)
+                            yield Completion(
+                                k_repr + "]",
+                                -len(key),
+                                display=f"[{k_repr}]",
+                                display_meta=self._do_repr(result[k]),
+                            )
+                        except ReprFailedError:
+                            pass
+
+            # Complete list/tuple index keys.
+            elif isinstance(result, (list, tuple)):
+                if not key or key.isdigit():
+                    for k in range(min(len(result), 1000)):
+                        if str(k).startswith(key):
+                            try:
+                                k_repr = self._do_repr(k)
+                                yield Completion(
+                                    k_repr + "]",
+                                    -len(key),
+                                    display=f"[{k_repr}]",
+                                    display_meta=self._do_repr(result[k]),
+                                )
+                            except ReprFailedError:
+                                pass
+
+    def _get_attribute_completions(
+        self,
+        document: Document,
+        complete_event: CompleteEvent,
+        temp_locals: Dict[str, Any],
+    ) -> Iterable[Completion]:
+        """
+        Complete attribute names.
+        """
+        match = self.attribute_lookup_pattern.search(document.text_before_cursor)
+        if match is not None:
+            object_var, attr_name = match.groups()
+
+            # Do lookup of `object_var` in the context.
+            result = self._lookup(object_var, temp_locals)
+
+            names = self._sort_attribute_names(dir(result))
+
+            for name in names:
+                if name.startswith(attr_name):
+                    yield Completion(
+                        name, -len(attr_name),
+                    )
+
+    def _sort_attribute_names(self, names: List[str]) -> List[str]:
+        """
+        Sort attribute names alphabetically, but move the double underscore and
+        underscore names to the end.
+        """
+
+        def sort_key(name: str):
+            if name.startswith("__"):
+                return (2, name)  # Double underscore comes latest.
+            if name.startswith("_"):
+                return (1, name)  # Single underscore before that.
+            return (0, name)  # Other names first.
+
+        return sorted(names, key=sort_key)
+
+
+class ReprFailedError(Exception):
+    " Raised when the repr() call in `DictionaryCompleter` fails. "
 
 
 try:
